import os
import torch
from torch.utils.data import Dataset
import numpy as np
import csv
import random
import dgl
import itertools
import pickle
from dgl.data.utils import load_graphs, save_graphs
from tqdm import tqdm


class cl_subgraphs(Dataset):
	def __init__(self, root, args, feat,
	             num_labels_classes):
		self.feat = feat
		self.sample_nodes = args.sample_nodes
		self.num_labels_classes = num_labels_classes   
		self.input_x, _label_dict = dgl.data.utils.load_graphs(root + '/processed/input_x.bin')
		self.input_center = np.load(root + '/processed/input_center.npy')
		self.total_ndata_h = np.load(root + '/processed/total_ndata_h.npy', allow_pickle=True).tolist()
	
	def loadCSV(self, csvf):
		dictGraphsLabels = {}
		dictLabels = {}
		dictGraphs = {}
		
		with open(csvf) as csvfile:
			csvreader = csv.reader(csvfile, delimiter=',')
			next(csvreader, None)   
			for i, row in enumerate(csvreader):
				filename = row[1]
				g_idx = int(filename.split('_')[0])
				label = row[2]
				 
				
				if g_idx in dictGraphs.keys():
					dictGraphs[g_idx].append(filename)
				else:
					dictGraphs[g_idx] = [filename]
					dictGraphsLabels[g_idx] = {}
				
				if label in dictGraphsLabels[g_idx].keys():
					dictGraphsLabels[g_idx][label].append(filename)
				else:
					dictGraphsLabels[g_idx][label] = [filename]
				
				if label in dictLabels.keys():
					dictLabels[label].append(filename)
				else:
					dictLabels[label] = [filename]
		return dictLabels, dictGraphs, dictGraphsLabels
	
	def create_ssl_data(self):
		"""
		create the entire set of batches of tasks for disjoint label setting, independent of  
		"""
		self.x = self.data[:][:]
		if self.link_pred_mode:
			info = [self.generate_subgraph_link_pred(self.G[int(item.split('_')[0])], int(item.split('_')[1]),
			                                         int(item.split('_')[2]), item)
			        for sublist in self.x for item in sublist]
		else:
			info = [self.generate_subgraph(self.G[int(item.split('_')[0])], int(item.split('_')[1]), item)
			        for sublist in self.x for item in sublist]
		input_graph_idx = [int(item.split('_')[0])   
		                   for sublist in self.x for item in sublist]
		
		input_x = [i for i, j, k in info]
		input_y = np.array([self.subgraph2label[item]
		                    for sublist in self.x for item in sublist]).astype(np.int32)
		
		input_center = np.array([j for i, j, k in info]).astype(np.int32)
		input_node_idx = [k for i, j, k in info]
		 
		
		return input_x, torch.LongTensor(input_y), input_center, input_node_idx, input_graph_idx
	
	 
	def generate_subgraph(self, G, i, item):
		if item in self.subgraphs:
			return self.subgraphs[item]
		else:
			 
			if self.h == 2:
				f_hop = [n.item() for n in G.in_edges(i)[0]]
				n_l = [[n.item() for n in G.in_edges(i)[0]] for i in f_hop]
				h_hops_neighbor = torch.tensor(list(set(list(itertools.chain(*n_l)) + f_hop + [i]))).numpy()
			elif self.h == 1:
				f_hop = [n.item() for n in G.in_edges(i)[0]]
				h_hops_neighbor = torch.tensor(list(set(f_hop + [i]))).numpy()
			elif self.h == 3:
				f_hop = [n.item() for n in G.in_edges(i)[0]]
				n_2 = [[n.item() for n in G.in_edges(i)[0]] for i in f_hop]
				n_3 = [[n.item() for n in G.in_edges(i)[0]] for i in list(itertools.chain(*n_2))]
				h_hops_neighbor = torch.tensor(
					list(set(list(itertools.chain(*n_2)) + list(itertools.chain(*n_3)) + f_hop + [i]))).numpy()
			if h_hops_neighbor.reshape(-1, ).shape[0] > self.sample_nodes:
				h_hops_neighbor = np.random.choice(h_hops_neighbor, self.sample_nodes, replace=False)
				h_hops_neighbor = np.unique(np.append(h_hops_neighbor, [i]))
			
			sub = G.subgraph(h_hops_neighbor)
			h_c = list(sub.ndata[dgl.NID].numpy())
			dict_ = dict(zip(h_c, list(range(len(h_c)))))
			self.subgraphs[item] = (sub, dict_[i], h_c)
			
			return sub, dict_[i], h_c
	
	def __getitem__(self, index):
		 
		 
		 
		 
		
		 
		
		 
		 
		self.input_x[index].ndata['h'] = torch.from_numpy(
			self.total_ndata_h[index])   
		 
		
		input_x_0, input_x_1 = graph_aug2view(self.input_x[index])
		return input_x_0, input_x_1   
	
	 
	
	def __len__(self):
		 
		return len(self.input_x)


class l2l_subgraphs(cl_subgraphs):
	def __init__(self, root, args, feat,
	             num_labels_classes):
		super(l2l_subgraphs, self).__init__(root, args, feat, num_labels_classes)
	
	def __getitem__(self, index):
		self.input_x[index].ndata['h'] = torch.from_numpy(self.total_ndata_h[index])
		
		input_x_0, input_x_1 = graph_aug2view(self.input_x[index])
		return input_x_0, input_x_1, self.input_center[index]


class l2l_other_subgraphs(cl_subgraphs):
	def __init__(self, root, args, feat,
	             num_labels_classes):
		 
		self.feat = feat
		self.sample_nodes = 1000
		self.num_labels_classes = num_labels_classes
		
		self.input_x, _label_dict = dgl.data.utils.load_graphs(root + '/processed/input_x.bin')
		self.input_center = np.load(root + 'processed/input_center.npy', allow_pickle=True).tolist()
		self.input_graph_idx = np.load(root + 'processed/input_graph_idx.npy', allow_pickle=True).tolist()
		self.input_node_idx = np.load(root + 'processed/input_node_idx.npy', allow_pickle=True).tolist()
		self.input_node_idx = [torch.Tensor(node_i).long() for node_i in
		                       self.input_node_idx]   
		self.feat = [torch.Tensor(feat_i.astype(float)) for feat_i in self.feat]
	
	def __getitem__(self, index):
		attr = self.feat[self.input_graph_idx[index]][self.input_node_idx[index]]
		self.input_x[index].ndata['h'] = attr
		input_x_0, input_x_1 = graph_aug2view(self.input_x[index])
		return input_x_0, input_x_1, self.input_center[index]


def cl_collate(samples):
	 
	 
	 
	input_x_0, input_x_1 = map(list, zip(*samples))
	
	input_x_0 = dgl.batch(input_x_0)
	input_x_1 = dgl.batch(input_x_1)
	return input_x_0, input_x_1   


def l2l_collate(samples):
	 
	 
	 
	input_x_0, input_x_1, input_center = map(list, zip(*samples))
	
	input_x_0 = dgl.batch(input_x_0)
	input_x_1 = dgl.batch(input_x_1)
	return input_x_0, input_x_1, torch.LongTensor(
		input_center)   


def cka_cl_collate(samples):
	input_x = map(list, zip(*samples))
	input_x = dgl.batch(input_x)
	return input_x


 
 
 
 


def drop_edge(graph, drop_prob):
	E = graph.num_edges()
	if E <= 1:
		return graph
	mask_rates = torch.FloatTensor(np.ones(E) * drop_prob)
	masks = torch.bernoulli(1 - mask_rates)
	index = random.randint(0, len(masks) - 1)
	masks[index] = 1
	edge_idx = masks.nonzero().squeeze(1)
	
	sg = dgl.edge_subgraph(graph, edge_idx, relabel_nodes=False)   
	 
	
	return sg


def drop_node(graph, drop_prob):
	N = graph.num_nodes()
	
	mask_rates = torch.FloatTensor(np.ones(N) * drop_prob)
	masks = torch.bernoulli(1 - mask_rates)
	index = random.randint(0, len(masks) - 1)
	masks[index] = 1
	node_idx = masks.nonzero().squeeze(1)
	sg = dgl.node_subgraph(graph, node_idx)
	 
	
	return sg


def drop_feat(x, drop_prob):
	D = x.shape[1]
	mask_rates = torch.FloatTensor(np.ones(D) * drop_prob)
	masks = torch.bernoulli(mask_rates)
	x[:, masks.bool()] = 0
	
	return x


def graph_aug(graph):
	 
	graph = drop_edge(graph, 0.3)   
	graph.ndata['h'] = drop_feat(graph.ndata['h'], 0.35)
	
	return graph


def graph_aug2view(graph):
	view0 = dgl.add_self_loop(graph_aug(graph))
	view1 = dgl.add_self_loop(graph_aug(graph))
	return view0, view1


if __name__ == '__main__':
	with open('/home/ name/data/fold_PPI/label.pkl', 'rb') as f:
		info = pickle.load(f)
	
	feat = np.load('/home/ name/data/fold_PPI/features.npy', allow_pickle=True)
	if len(feat.shape) == 2:
		print('single graph, to make it compatible to multiple graph retrieval.')
		feat = [feat]
	
	dgl_graph, _label_dict = load_graphs("/home/ name/data/fold_PPI/graph_dgl.bin")
	a = cl_subgraphs(root='/home/ name/data/fold_PPI/', mode='train', subgraph2label=info, args=None, adjs=dgl_graph,
	                 h=2, feat=feat, n_way=3, k_shot=3, k_query=10, batchsz=1000, )
