import torch, os

torch.multiprocessing.set_sharing_strategy('file_system')
import sys

sys.path.append("..")
import torch.multiprocessing

curPath = os.path.abspath(os.path.dirname(__file__))
rootPath = os.path.split(curPath)[0]
sys.path.append(rootPath)

from gmeta.subgraph_data_processing import Subgraphs, collate, Subgraphs4test
from torch.utils.data import DataLoader
import random, sys, pickle
import argparse
from dgl.data.utils import load_graphs, save_graphs
import gmeta.gm_meta
import numpy as np

from gmeta.meta import cl_meta
from gmeta.gnn_zoo import GCN
import time
import copy
import psutil
from memory_profiler import memory_usage
from utils import save_checkpoint, console_log, load_pretrained, SubsetRandomSampler
import torch.nn.functional as F


def main():
	mem_usage = memory_usage(-1, interval=.5, timeout=1)
	torch.manual_seed(0)
	torch.cuda.manual_seed_all(0)
	np.random.seed(0)
	random.seed(0)
	
	print(str(args))
	
	root = args.data_dir
	
	feat = np.load(root + 'features.npy', allow_pickle=True)
	
	dgl_graph, _label_dict = load_graphs(root + "/graph_dgl.bin")
	
	if args.task_setup == 'Disjoint':
		with open(root + 'label.pkl', 'rb') as f:
			info = pickle.load(f)
	elif args.task_setup == 'Shared':
		if args.task_mode == 'True':
			root = root + '/task' + str(args.task_n) + '/'
		with open(root + 'label.pkl', 'rb') as f:
			info = pickle.load(f)
	
	total_class = len(np.unique(np.array(list(info.values()))))
	 
	print('There are {} classes '.format(total_class))
	
	if args.task_setup == 'Disjoint':
		print('task_setup is Disjoint')
		labels_num = args.n_way
	elif args.task_setup == 'Shared':
		print('task_setup is Shared')
		labels_num = total_class
	
	if len(feat.shape) == 2:
		 
		feat = [feat]
	args.n_feats = feat[0].shape[1]
	
	config = []
	
	if args.link_pred_mode == 'True':
		config.append(('LinkPred', [True]))
	
	db_train = Subgraphs(root, 'train', info, n_way=args.n_way, k_shot=args.k_spt, k_query=args.k_qry,
	                     batchsz=args.batchsz, args=args, adjs=dgl_graph, h=args.h, feat=feat)
	db_val = Subgraphs(root, 'val', info, n_way=args.n_way, k_shot=args.k_spt, k_query=args.k_qry, batchsz=100,
	                   args=args, adjs=dgl_graph, h=args.h, feat=feat)
	db_test = Subgraphs4test(root, 'test', info, n_way=args.n_way, k_shot=args.k_spt, k_query=args.k_qry, batchsz=100,
	                         args=args, adjs=dgl_graph, h=args.h, feat=feat)
	 
	device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
	args.net = GCN(in_feats=args.n_feats,
	               n_hidden=args.n_hidden,
	               n_classes=args.n_classes,
	               n_layers=args.n_layers,
	               activation=args.activation,
	               dropout=args.dropout,
	               batchnorm=args.batchnorm,
	               ssl=False,
	               residual=args.residual,
	               LinkPred_mode=args.link_pred_mode)
	args.net = args.net.to(device)
	 
	
	if args.ft_mlp:
		print('only ft MLP.')
		linear_keyword = 'readout'   
		for name, param in args.net.named_parameters():
			if name not in ['%s.weight' % linear_keyword, '%s.bias' % linear_keyword]:
				print(name, ' no grad;')
				param.requires_grad = False
		
		maml = gmeta.gm_meta._cl_meta(args, config)
		maml = maml.to(device)
	else:
		print('ft whole gnn.')
		maml = cl_meta(args, config)
		maml = maml.to(device)
	
	tmp = filter(lambda x: x.requires_grad, maml.parameters())
	num = sum(map(lambda x: np.prod(x.shape), tmp))
	print(str(maml))
	print('Total trainable tensors:', num)
	
	max_acc = 0
	model_max = None
	
	print('------ Start Training ------')
	s_start = time.time()
	max_memory = 0
	db_t = DataLoader(db_test, 1, shuffle=True, num_workers=args.num_workers, pin_memory=True, collate_fn=collate,
	                  persistent_workers=True)
	 
	 
	label_rate = args.label_rate
	num_train = len(db_train)
	indices = list(range(num_train))
	split = int(np.floor((1.0-label_rate) * num_train))
	
	np.random.seed(0)
	np.random.shuffle(indices)
	
	train_idx, valid_idx = indices[split:], indices[:split]
	sampler_train = SubsetRandomSampler(train_idx)
	sampler_valid = SubsetRandomSampler(valid_idx)
	accs_all_test = []
	if args.finetune:
		 
		db = DataLoader(db_train, args.task_num, sampler=sampler_train, num_workers=args.num_workers, pin_memory=True, collate_fn=collate, persistent_workers=True)
		db_v = DataLoader(db_val, 1, shuffle=True, num_workers=args.num_workers, pin_memory=True, collate_fn=collate,
		                  persistent_workers=True)
		 
		 
		for epoch in range(args.epoch):
			s_f = time.time()
			for step, (x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry) in enumerate(db):
				 
				 
				nodes_len = 0
				if step >= 1:
					data_loading_time = time.time() - s_r
				else:
					data_loading_time = time.time() - s_f
				s = time.time()
				 
				 
				nodes_len += sum([sum([len(j) for j in i]) for i in n_spt])
				 
				accs = maml(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat)
				max_memory = max(max_memory, float(psutil.virtual_memory().used / (1024 ** 3)))
				if step % args.train_result_report_steps == 0:
					print('Epoch:', epoch + 1, ' Step:', step, ' training acc:', str(accs[-1])[:5], ' time elapsed:',
					      str(time.time() - s)[:5], ' data loading takes:', str(data_loading_time)[:5],
					      ' Memory usage:',
					      str(float(psutil.virtual_memory().used / (1024 ** 3)))[:5])
				s_r = time.time()
			
			 
			accs_all_test = []
			
			for x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry in db_v:
				accs = maml.finetunning(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat)
				accs_all_test.append(accs)
			
			accs = np.array(accs_all_test).mean(axis=0).astype(np.float16)
			print('Epoch:', epoch + 1, ' Val acc:', str(accs[-1])[:5])
			if accs[-1] > max_acc:
				max_acc = accs[-1]
				model_max = copy.deepcopy(maml.net)
		
		for x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry in db_t:
			accs = maml.finetunning(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat)
			accs_all_test.append(accs)
		
		accs = np.array(accs_all_test).mean(axis=0).astype(np.float16)
		print('Test acc:', str(accs[1])[:5])
		maml.net = model_max
	
	for x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry in db_t:
		accs = maml.finetunning(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat)
		accs_all_test.append(accs)
	
	accs = np.array(accs_all_test).mean(axis=0).astype(np.float16)
	print('Early Stopped Test acc:', str(accs[-1])[:5])
	print('Total Time:', str(time.time() - s_start)[:5])
	print('Max Momory:', str(max_memory)[:5])


if __name__ == '__main__':
	argparser = argparse.ArgumentParser()
	argparser.add_argument('--outdir', default='outputs/', type=str,
	                       help='path to location where to save outputs. Empty for current working dir')
	argparser.add_argument('--exp_name', default='bgrl', type=str, help='name under outputs/')
	argparser.add_argument('--use_amp', action='store_true', help='whether use amp')
	argparser.add_argument('--epoch', type=int, help='epoch number', default=10)
	argparser.add_argument('--n_way', type=int, help='n way', default=3)
	argparser.add_argument('--k_spt', type=int, help='k shot for support set', default=3)
	argparser.add_argument('--k_qry', type=int, help='k shot for query set', default=24)
	argparser.add_argument('--label_rate', type=float, default=1.0, help="the label_rate for finetunning")
	argparser.add_argument('--task_num', type=int, help='meta batch size, namely task num', default=32)
	argparser.add_argument('--meta_lr', type=float, help='meta-level outer learning rate', default=1e-3)
	 
	argparser.add_argument('--update_lr', type=float, help='task-level inner update learning rate', default=0.01)
	argparser.add_argument('--update_step', type=int, help='task-level inner update steps', default=10)
	argparser.add_argument('--update_step_test', type=int, help='update steps for finetunning', default=20)
	argparser.add_argument('--input_dim', type=int, help='input feature dim', default=1)
	argparser.add_argument('--attention_size', type=int, help='dim of attention_size', default=32)
	argparser.add_argument("--data_dir", default='/home/ name/data/arxiv/', type=str, required=False,
	                       help="The input data dir.")
	argparser.add_argument("--ckpt_path",
	                       default='/home/ name/research/G-Meta/outputs/bgrl/fold_0/bs2048-warmup5-lr0.05-m0.99-layer1/ckpt.pth.tar',
	                       type=str, required=False, help="cl pretrained ckpt dir.")
	argparser.add_argument("--task_setup", default='Disjoint', type=str, required=False,
	                       help="Select from Disjoint or Shared Setup. For Disjoint-Label, single/multiple graphs are both considered.")
	argparser.add_argument("--num_workers", default=2, type=int, required=False, help="num of workers")
	argparser.add_argument("--batchsz", default=1000, type=int, required=False, help="batch size")
	argparser.add_argument("--link_pred_mode", type=bool, default=False, required=False, help="For Link Prediction")
	argparser.add_argument("--h", default=2, type=int, required=False, help="neighborhood size")
	argparser.add_argument("--maml", default=True, type=bool, required=False, help="whether use maml finetune.")
	argparser.add_argument("--finetune", default=True, type=bool, help="finetune on FULL dataset.")
	argparser.add_argument("--net", help="net used in maml finetune.")
	argparser.add_argument("--method", default='gmeta', type=str, required=False, help="Use gmeta")
	argparser.add_argument('--task_n', type=int, help='task number', default=1)
	argparser.add_argument("--task_mode", default='False', type=str, required=False, help="For Evaluating on Tasks")
	argparser.add_argument("--val_result_report_steps", default=100, type=int, required=False, help="validation report")
	argparser.add_argument("--train_result_report_steps", default=30, type=int, required=False, help="training report")
	argparser.add_argument('--sample_nodes', type=int, help='sample nodes if above this number of nodes', default=1000)
	
	argparser.add_argument('--n_layers', type=int, default=1, help="number of hidden layers")
	argparser.add_argument('--n_feats', type=int, help="number of features[1]")
	argparser.add_argument('--n_hidden', type=int, default=256, help="number of hidden units")
	argparser.add_argument('--n_classes', type=int, default=3, help="number of classes, equal to GNN output dimension")
	argparser.add_argument('--in_drop', type=float, default=0, help="input feature dropout")
	argparser.add_argument('--attn_drop', type=float, default=0, help="attention dropout")
	argparser.add_argument('--negative_slope', type=float, default=0.2, help="the negative slope of leaky relu")
	argparser.add_argument('--num_heads', type=int, default=8, help="number of hidden attention heads")
	argparser.add_argument('--num_out_heads', type=int, default=1, help="number of output attention heads")
	argparser.add_argument('--heads', type=int, default=1, help="number of all heads in GAT")
	argparser.add_argument('--activation', help="activations (F.elu) in GAT")
	argparser.add_argument('--dropout', type=float, default=0.2, help="dropout rate for GNN")
	argparser.add_argument('--ft_mlp', type=bool, default=False, help="only MLP readout optimized")
	argparser.add_argument('--batchnorm', type=bool, default=True, help="use batchnorm for GNN")
	argparser.add_argument('--residual', type=bool, default=False, help="use residual for gnn")
	argparser.add_argument('-p', '--print_freq', default=5, type=int, metavar='N', help='print frequency (default: 10)')
	argparser.add_argument('--norm', action='store_true', default=False, help="use F.normalize in GAT output")
	
	args = argparser.parse_args()
	args.activation = F.leaky_relu   
	args.outdir = os.path.join(args.outdir, args.exp_name,
	                           'lr{lr}-h{h}-layer{n_layers}-n_cls{n_classes}'.format(
		                           h=args.h, n_layers=args.n_layers, n_classes=args.n_classes, lr=args.update_lr))
	
	if os.path.exists(args.outdir) is None:
		os.makedirs(os.path.join(args.outdir), exist_ok=True)
	
	console = console_log(args.outdir)
	main()
 
 
