import  torch

import os
torch.multiprocessing.set_sharing_strategy('file_system')
import sys
sys.path.append("..")
import torch.multiprocessing
curPath = os.path.abspath(os.path.dirname(__file__))
rootPath = os.path.split(curPath)[0]
sys.path.append(rootPath)

from    torch import nn
from    torch.nn import functional as F
from    torch.utils.data import TensorDataset, DataLoader
from    torch import optim
import  numpy as np
from    gmeta.learner import Classifier
from    copy import copy
from learn2learn.utils import clone_module, update_module
from copy import deepcopy
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def euclidean_dist(x, y):
     
     
    n = x.size(0)
    m = y.size(0)
    d = x.size(1)
    if d != y.size(1):
        raise Exception

    x = x.unsqueeze(1).expand(n, m, d)
    y = y.unsqueeze(0).expand(n, m, d)

    return torch.pow(x - y, 2).sum(2)

def proto_loss_spt(logits, y_t, n_support):
    target_cpu = y_t.to('cpu')
    input_cpu = logits.to('cpu')
    
    def supp_idxs(c):
        return target_cpu.eq(c).nonzero()[:n_support].squeeze(1)

    classes = torch.unique(target_cpu)
    n_classes = len(classes)
    n_query = n_support

    support_idxs = list(map(supp_idxs, classes))

    prototypes = torch.stack([input_cpu[idx_list].mean(0) for idx_list in support_idxs])
    query_idxs = torch.stack(list(map(lambda c: target_cpu.eq(c).nonzero()[:n_support], classes))).view(-1)
    query_samples = input_cpu[query_idxs]   
    dists = euclidean_dist(query_samples, prototypes)
    log_p_y = F.log_softmax(-dists, dim=1).view(n_classes, n_query, -1)

    target_inds = torch.arange(0, n_classes)
    target_inds = target_inds.view(n_classes, 1, 1)
    target_inds = target_inds.expand(n_classes, n_query, 1).long()

    loss_val = -log_p_y.gather(2, target_inds).squeeze().view(-1).mean()
    _, y_hat = log_p_y.max(2)
    acc_val = y_hat.eq(target_inds.squeeze()).float().mean()
    return loss_val, acc_val, prototypes

def proto_loss_qry(logits, y_t, prototypes):
    target_cpu = y_t.to('cpu')
    input_cpu = logits.to('cpu')

    classes = torch.unique(target_cpu)
    n_classes = len(classes)

    n_query = int(logits.shape[0]/n_classes)

    query_idxs = torch.stack(list(map(lambda c: target_cpu.eq(c).nonzero(), classes))).view(-1)
    query_samples = input_cpu[query_idxs]

    dists = euclidean_dist(query_samples, prototypes)

    log_p_y = F.log_softmax(-dists, dim=1).view(n_classes, n_query, -1)

    target_inds = torch.arange(0, n_classes)
    target_inds = target_inds.view(n_classes, 1, 1)
    target_inds = target_inds.expand(n_classes, n_query, 1).long()

    loss_val = -log_p_y.gather(2, target_inds).squeeze().view(-1).mean()
    _, y_hat = log_p_y.max(2)
    acc_val = y_hat.eq(target_inds.squeeze()).float().mean()
    return loss_val, acc_val


class GMMeta(nn.Module):
    def __init__(self, args, config):
        super(GMMeta, self).__init__()
        self.update_lr = args.update_lr
        self.meta_lr = args.meta_lr
        self.n_way = args.n_way
        self.k_spt = args.k_spt
        self.k_qry = args.k_qry
        self.task_num = args.task_num
        self.update_step = args.update_step
        self.update_step_test = args.update_step_test
        
        self.net = Classifier(config)
        self.net = self.net.to(device)
        
        self.meta_optim = optim.Adam(self.net.parameters(), lr=self.meta_lr)
        
        self.method = args.method
    
    def forward_ProtoMAML(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        """
        b: number of tasks
        setsz: the size for each task
        :param x_spt:   [b], where each unit is a mini-batch of subgraphs, i.e. x_spt[0] is a DGL batch of  
        :param y_spt:   [b, setsz]
        :param x_qry:   [b], where each unit is a mini-batch of subgraphs, i.e. x_spt[0] is a DGL batch of  
        :param y_qry:   [b, querysz]
        :return:
        """
        task_num = len(x_spt)
        querysz = len(y_qry[0])
        losses_s = [0 for _ in range(self.update_step)]
        losses_q = [0 for _ in range(self.update_step + 1)]   
        corrects = [0 for _ in range(self.update_step + 1)]
        
        for i in range(task_num):
            feat_spt = torch.Tensor(np.vstack(([feat[g_spt[i][j]][np.array(x)] for j, x in enumerate(n_spt[i])]))).to(
                device)
            feat_qry = torch.Tensor(np.vstack(([feat[g_qry[i][j]][np.array(x)] for j, x in enumerate(n_qry[i])]))).to(
                device)
             
            logits, _ = self.net(x_spt[i].to(device), c_spt[i].to(device), feat_spt, vars=None)
            loss, _, prototypes = proto_loss_spt(logits, y_spt[i], self.k_spt)
            losses_s[0] += loss
            grad = torch.autograd.grad(loss, self.net.parameters())
            fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, self.net.parameters())))
            
             
            with torch.no_grad():
                 
                logits_q, _ = self.net(x_qry[i].to(device), c_qry[i].to(device), feat_qry, self.net.parameters())
                 
                y_qry[i] = y_qry[i].to(device)
                loss_q = F.cross_entropy(logits_q,
                                         y_qry[i])   
                losses_q[0] += loss_q
                
                pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                correct = torch.eq(pred_q, y_qry[i]).sum().item()   
                corrects[0] = corrects[0] + correct   
            
             
            with torch.no_grad():
                logits_q, _ = self.net(x_qry[i].to(device), c_qry[i].to(device), feat_qry, fast_weights)
                y_qry[i] = y_qry[i].to(device)
                loss_q = F.cross_entropy(logits_q,
                                         y_qry[i])   
                losses_q[1] += loss_q
                
                pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                correct = torch.eq(pred_q, y_qry[i]).sum().item()
                corrects[1] = corrects[1] + correct   
            
            for k in range(1, self.update_step):
                 
                logits, _ = self.net(x_spt[i].to(device), c_spt[i].to(device), feat_spt, fast_weights)
                y_qry[i], y_spt[i] = y_qry[i].to(device), y_spt[i].to(device)
                loss = F.cross_entropy(logits,
                                       y_spt[i])   
                losses_s[k] += loss
                 
                grad = torch.autograd.grad(loss, fast_weights, retain_graph=True)
                 
                fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, fast_weights)))
                logits_q, _ = self.net(x_qry[i].to(device), c_qry[i].to(device), feat_qry, fast_weights)
                 
                loss_q = F.cross_entropy(logits_q,
                                         y_qry[i])   
                losses_q[k + 1] += loss_q
                
                with torch.no_grad():
                    pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                    correct = torch.eq(pred_q, y_qry[i]).sum().item()   
                    corrects[k + 1] = corrects[k + 1] + correct
        
         
         
        loss_q = losses_q[-1] / task_num
        
        if torch.isnan(loss_q):
            pass
        else:
             
            self.meta_optim.zero_grad()
            loss_q.backward()
            self.meta_optim.step()
        
        accs = np.array(corrects) / (task_num)
        
        return accs
    
    def _forward_ProtoMAML(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        """
        b: number of tasks
        setsz: the size for each task
        :param x_spt:   [b], where each unit is a mini-batch of subgraphs, i.e. x_spt[0] is a DGL batch of  
        :param y_spt:   [b, setsz]
        :param x_qry:   [b], where each unit is a mini-batch of subgraphs, i.e. x_spt[0] is a DGL batch of  
        :param y_qry:   [b, querysz]
        :return:
        """
        task_num = len(x_spt)
        querysz = len(y_qry[0])
        losses_s = [0 for _ in range(self.update_step)]
        losses_q = [0 for _ in range(self.update_step + 1)]   
        corrects = [0 for _ in range(self.update_step + 1)]
        
        for i in range(task_num):
            feat_spt = torch.Tensor(np.vstack(([feat[g_spt[i][j]][np.array(x)] for j, x in enumerate(n_spt[i])]))).to(
                device)
            feat_qry = torch.Tensor(np.vstack(([feat[g_qry[i][j]][np.array(x)] for j, x in enumerate(n_qry[i])]))).to(
                device)
             
            logits, _ = self.net(x_spt[i].to(device), c_spt[i].to(device), feat_spt, vars=None)
            loss, _, prototypes = proto_loss_spt(logits, y_spt[i], self.k_spt)
            losses_s[0] += loss
            grad = torch.autograd.grad(loss, self.net.parameters())
            fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, self.net.parameters())))
            
             
            with torch.no_grad():
                 
                logits_q, _ = self.net(x_qry[i].to(device), c_qry[i].to(device), feat_qry, self.net.parameters())
                loss_q, acc_q = proto_loss_qry(logits_q, y_qry[i], prototypes)
                losses_q[0] += loss_q
                corrects[0] = corrects[0] + acc_q
            
             
            with torch.no_grad():
                logits_q, _ = self.net(x_qry[i].to(device), c_qry[i].to(device), feat_qry, fast_weights)
                loss_q, acc_q = proto_loss_qry(logits_q, y_qry[i], prototypes)
                losses_q[1] += loss_q
                corrects[1] = corrects[1] + acc_q
            
            for k in range(1, self.update_step):
                 
                logits, _ = self.net(x_spt[i].to(device), c_spt[i].to(device), feat_spt, fast_weights)
                loss, _, prototypes = proto_loss_spt(logits, y_spt[i], self.k_spt)
                losses_s[k] += loss
                 
                grad = torch.autograd.grad(loss, fast_weights, retain_graph=True)
                 
                fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, fast_weights)))
                logits_q, _ = self.net(x_qry[i].to(device), c_qry[i].to(device), feat_qry, fast_weights)
                 
                loss_q, acc_q = proto_loss_qry(logits_q, y_qry[i], prototypes)
                losses_q[k + 1] += loss_q
                
                corrects[k + 1] = corrects[k + 1] + acc_q
        
         
         
        loss_q = losses_q[-1] / task_num
        
        if torch.isnan(loss_q):
            pass
        else:
             
            self.meta_optim.zero_grad()
            loss_q.backward()
            self.meta_optim.step()
        
        accs = np.array(corrects) / (task_num)
        
        return accs
    
    def finetunning_ProtoMAML(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        querysz = len(y_qry[0])
        
        corrects = [0 for _ in range(self.update_step_test + 1)]
        
         
        net = deepcopy(self.net)
        x_spt = x_spt[0]
        y_spt = y_spt[0]
        x_qry = x_qry[0]
        y_qry = y_qry[0]
        c_spt = c_spt[0]
        c_qry = c_qry[0]
        n_spt = n_spt[0]
        n_qry = n_qry[0]
        g_spt = g_spt[0]
        g_qry = g_qry[0]
        
        feat_spt = torch.Tensor(np.vstack(([feat[g_spt[j]][np.array(x)] for j, x in enumerate(n_spt)]))).to(device)
        feat_qry = torch.Tensor(np.vstack(([feat[g_qry[j]][np.array(x)] for j, x in enumerate(n_qry)]))).to(device)
        
         
        logits, _ = net(x_spt.to(device), c_spt.to(device), feat_spt)
        y_spt = y_spt.to(device)
        loss = F.cross_entropy(logits, y_spt)   
        grad = torch.autograd.grad(loss, net.parameters())
        fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, net.parameters())))
        
         
        with torch.no_grad():
             
            logits_q, _ = net(x_qry.to(device), c_qry.to(device), feat_qry, net.parameters())
            pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
             
            y_qry = y_qry.to(device)
            loss_q = F.cross_entropy(logits_q, y_qry)   
            correct = torch.eq(pred_q, y_qry).sum().item()
            corrects[0] = corrects[0] + correct
             
         
        with torch.no_grad():
             
            logits_q, _ = net(x_qry.to(device), c_qry.to(device), feat_qry, fast_weights)
            pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
             
            correct = torch.eq(pred_q, y_qry).sum().item()
            corrects[1] = corrects[1] + correct
             
             
        
        for k in range(1, self.update_step_test):
             
            logits, _ = net(x_spt.to(device), c_spt.to(device), feat_spt, fast_weights)
            loss = F.cross_entropy(logits, y_spt)   
             
            grad = torch.autograd.grad(loss, fast_weights)   
             
            fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, fast_weights)))
            
            logits_q, _ = net(x_qry.to(device), c_qry.to(device), feat_qry, fast_weights)
             
            loss_q = F.cross_entropy(logits_q, y_qry)   
            with torch.no_grad():
                pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                correct = torch.eq(pred_q, y_qry).sum().item()   
                corrects[k + 1] = corrects[k + 1] + correct
             
        
        del net
        accs = np.array(corrects)
        
        return accs
    
    def _finetunning_ProtoMAML(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        querysz = len(y_qry[0])
        
        corrects = [0 for _ in range(self.update_step_test + 1)]
        
         
        net = deepcopy(self.net)
        x_spt = x_spt[0]
        y_spt = y_spt[0]
        x_qry = x_qry[0]
        y_qry = y_qry[0]
        c_spt = c_spt[0]
        c_qry = c_qry[0]
        n_spt = n_spt[0]
        n_qry = n_qry[0]
        g_spt = g_spt[0]
        g_qry = g_qry[0]
        
        feat_spt = torch.Tensor(np.vstack(([feat[g_spt[j]][np.array(x)] for j, x in enumerate(n_spt)]))).to(device)
        feat_qry = torch.Tensor(np.vstack(([feat[g_qry[j]][np.array(x)] for j, x in enumerate(n_qry)]))).to(device)
        
         
        logits, _ = net(x_spt.to(device), c_spt.to(device), feat_spt)
        loss, _, prototypes = proto_loss_spt(logits, y_spt, self.k_spt)
        grad = torch.autograd.grad(loss, net.parameters())
        fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, net.parameters())))
        
         
        with torch.no_grad():
             
            logits_q, _ = net(x_qry.to(device), c_qry.to(device), feat_qry, net.parameters())
            loss_q, acc_q = proto_loss_qry(logits_q, y_qry, prototypes)
            corrects[0] = corrects[0] + acc_q
         
        with torch.no_grad():
             
            logits_q, _ = net(x_qry.to(device), c_qry.to(device), feat_qry, fast_weights)
            loss_q, acc_q = proto_loss_qry(logits_q, y_qry, prototypes)
            corrects[1] = corrects[1] + acc_q
        
        for k in range(1, self.update_step_test):
             
            logits, _ = net(x_spt.to(device), c_spt.to(device), feat_spt, fast_weights)
            loss, _, prototypes = proto_loss_spt(logits, y_spt, self.k_spt)
             
            grad = torch.autograd.grad(loss, fast_weights, retain_graph=True)
             
            fast_weights = list(map(lambda p: p[1] - self.update_lr * p[0], zip(grad, fast_weights)))
            
            logits_q, _ = net(x_qry.to(device), c_qry.to(device), feat_qry, fast_weights)
             
            loss_q, acc_q = proto_loss_qry(logits_q, y_qry, prototypes)
            corrects[k + 1] = corrects[k + 1] + acc_q
        
        del net
        accs = np.array(corrects)
        
        return accs
    
    def forward(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        accs = self.forward_ProtoMAML(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat)
        return accs
    
    def finetunning(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        accs = self.finetunning_ProtoMAML(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry,
                                              feat)
        return accs


class MLPMeta(nn.Module):
    def __init__(self, args, config):
        super(MLPMeta, self).__init__()
    
    def forward_ProtoMAML(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
         
        """
        b: number of tasks
        setsz: the size for each task
        :param x_spt:   [b], where each unit is a mini-batch of subgraphs, i.e. x_spt[0] is a DGL batch of  
        :param y_spt:   [b, setsz]
        :param x_qry:   [b], where each unit is a mini-batch of subgraphs, i.e. x_spt[0] is a DGL batch of  
        :param y_qry:   [b, querysz]
        :return:
        """
        assert len(x_spt) == len(x_qry)
        task_num = len(x_qry)
        querysz = len(y_qry[0])
        losses_s = [0 for _ in range(self.update_step)]
        losses_q = [0 for _ in range(self.update_step + 1)]   
        corrects = [0 for _ in range(self.update_step + 1)]
        
        for i in range(task_num):
            feat_spt = torch.Tensor(np.vstack(([feat[g_spt[i][j]][np.array(x)] for j, x in enumerate(n_spt[i])]))).to(
                device)
            feat_qry = torch.Tensor(np.vstack(([feat[g_qry[i][j]][np.array(x)] for j, x in enumerate(n_qry[i])]))).to(
                device)
            x_spt[i] = x_spt[i].to(device)
             
            logits = self.net(x_spt[i], feat_spt, c_spt[i].to(device))
            loss, _, prototypes = proto_loss_spt(logits, y_spt[i], self.k_spt)
            losses_s[0] += loss
            filter_para = list(filter(lambda p: p.requires_grad, self.net.parameters()))
            grad = torch.autograd.grad(loss, filter_para)
             
            self.net_fast_weight = clone_module(self.net)
            filter_para_fast = list(filter(lambda p: p.requires_grad, self.net_fast_weight.parameters()))
            for _fast_param, _grad, _param in zip(filter_para, grad, filter_para_fast):
                _fast_param.data = _param.data - self.update_lr * _grad.data
                 
             
            with torch.no_grad():
                 
                x_qry[i] = x_qry[i].to(device)
                logits_q = self.net(x_qry[i], feat_qry, c_qry[i].to(device))
                y_qry[i] = y_qry[i].to(device)
                loss_q = F.cross_entropy(logits_q,
                                         y_qry[i])   
                losses_q[0] += loss_q
                
                pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                 
                correct = torch.eq(pred_q,
                                   y_qry[i]).float().mean().item()   
                corrects[0] = corrects[0] + correct   
            
             
            with torch.no_grad():
                 
                x_qry[i] = x_qry[i].to(device)
                logits_q = self.net_fast_weight(x_qry[i], feat_qry, c_qry[i].to(device))   
                y_qry[i] = y_qry[i].to(device)
                loss_q = F.cross_entropy(logits_q,
                                         y_qry[i])   
                losses_q[1] += loss_q
                
                pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                 
                correct = torch.eq(pred_q, y_qry[i]).float().mean().item()
                
                corrects[1] = corrects[1] + correct   
            
            for k in range(1, self.update_step):
                 
                 
                x_spt[i] = x_spt[i].to(device)
                logits = self.net_fast_weight(x_spt[i].to(device), feat_spt, c_spt[i].to(device))   
                y_qry[i], y_spt[i] = y_qry[i].to(device), y_spt[i].to(device)
                loss = F.cross_entropy(logits,
                                       y_spt[i])   
                losses_s[k] += loss
                 
                filter_para_fast = list(filter(lambda p: p.requires_grad, self.net.parameters()))
                grad = torch.autograd.grad(loss, filter_para_fast)   
                 
                 
                for _fast_param, _grad in zip(filter_para_fast, grad):
                    _fast_param.data = _fast_param.data - self.update_lr * _grad.data
                logits_q = self.net_fast_weight(x_qry[i].to(device), feat_qry, c_qry[i].to(device))   
                 
                loss_q = F.cross_entropy(logits_q,
                                         y_qry[i])   
                losses_q[k + 1] += loss_q
                
                with torch.no_grad():
                    pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                     
                    correct = torch.eq(pred_q, y_qry[i]).float().mean().item()   
                     
                    corrects[k + 1] = corrects[k + 1] + correct
        
         
         
        loss_q = losses_q[-1] / task_num
        
        if torch.isnan(loss_q):
            pass
        else:
             
            self.meta_optim.zero_grad()
            loss_q.backward()
            self.meta_optim.step()
        
        accs = np.array(corrects) / task_num   
        
        return accs
    
    def finetunning_ProtoMAML(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        querysz = len(y_qry[0])
        
        corrects = [0 for _ in range(self.update_step_test + 1)]
        
         
        net = clone_module(self.net)
        x_spt = x_spt[0]
        y_spt = y_spt[0]
        x_qry = x_qry[0]
        y_qry = y_qry[0]
        c_spt = c_spt[0]
        c_qry = c_qry[0]
        n_spt = n_spt[0]
        n_qry = n_qry[0]
        g_spt = g_spt[0]
        g_qry = g_qry[0]
        
        feat_spt = torch.Tensor(np.vstack(([feat[g_spt[j]][np.array(x)] for j, x in enumerate(n_spt)]))).to(device)
        feat_qry = torch.Tensor(np.vstack(([feat[g_qry[j]][np.array(x)] for j, x in enumerate(n_qry)]))).to(device)
        
         
        logits = net(x_spt.to(device), feat_spt, c_spt.to(device))
        y_spt = y_spt.to(device)
        loss = F.cross_entropy(logits, y_spt)   
        filter_para_wo_self = list(filter(lambda p: p.requires_grad, net.parameters()))
        grad = torch.autograd.grad(loss, filter_para_wo_self)
        
        self.net_fast_weight = clone_module(self.net)
        filter_para_fast = list(filter(lambda p: p.requires_grad, self.net_fast_weight.parameters()))
        
        for _fast_param, _grad, _param in zip(filter_para_fast, grad, filter_para_wo_self):
            _fast_param.data = _param.data - self.update_lr * _grad.data
             
         
        
         
        with torch.no_grad():
             
            logits_q = net(x_qry.to(device), feat_qry, c_qry.to(device))   
            pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
             
            y_qry = y_qry.to(device)
            loss_q = F.cross_entropy(logits_q, y_qry)   
             
            correct = torch.eq(pred_q, y_qry).float().mean().item()
            corrects[0] = corrects[0] + correct
             
         
        with torch.no_grad():
             
             
            logits_q = self.net_fast_weight(x_qry.to(device), feat_qry, c_qry.to(device))   
            pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
             
             
            correct = torch.eq(pred_q, y_qry).float().mean().item()
            
            corrects[1] = corrects[1] + correct
             
             
        
        for k in range(1, self.update_step_test):
             
            
            logits = self.net_fast_weight(x_spt.to(device), feat_spt, c_spt.to(device))   
            loss = F.cross_entropy(logits, y_spt)   
             
            filter_para_fast = list(filter(lambda p: p.requires_grad, self.net_fast_weight.parameters()))
            grad = torch.autograd.grad(loss, filter_para_fast)   
             
            for _fast_param, _grad in zip(filter_para_fast, grad):
                _fast_param.data = _fast_param.data - self.update_lr * _grad.data
             
            
            logits_q = self.net_fast_weight(x_qry.to(device), feat_qry, c_qry.to(device))   
             
            loss_q = F.cross_entropy(logits_q, y_qry)   
            with torch.no_grad():
                pred_q = F.softmax(logits_q, dim=1).argmax(dim=1)
                 
                correct = torch.eq(pred_q, y_qry).float().mean().item()   
                
                corrects[k + 1] = corrects[k + 1] + correct
             
        
        del net
        accs = np.array(corrects)   
        
        return accs
    
    def forward(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        accs = self.forward_ProtoMAML(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat)
        return accs
    
    def finetunning(self, x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry, feat):
        accs = self.finetunning_ProtoMAML(x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry,
                                              feat)
        return accs


class _cl_meta(MLPMeta):
     
    def __init__(self, args, config):
        super(_cl_meta, self).__init__(args, config)
        self.net = args.net
        
        self.update_lr = args.update_lr
        self.meta_lr = args.meta_lr
        self.n_way = args.n_way
        self.k_spt = args.k_spt
        self.k_qry = args.k_qry
        self.task_num = args.task_num
        self.update_step = args.update_step
        self.update_step_test = args.update_step_test
        
        self.net = self.net.to(device)
        self.net_fast_weight = clone_module(self.net)
        parameters = list(filter(lambda p: p.requires_grad, self.net.parameters()))
        self.meta_optim = optim.AdamW(parameters, lr=self.meta_lr)
        
        self.method = args.method