import torch
import torch.nn.functional as F
import dgl.function as fn
import torch.nn as nn
from torch.nn import init
import dgl
from dgl.nn.pytorch.conv.gatconv import GATConv
from dgl.nn.pytorch import AvgPooling

 
msg = fn.copy_src(src='h', out='m')
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

 
class GraphConv(nn.Module):
    def __init__(self,
                 in_feats,
                 out_feats,
                 activation=None):
        super(GraphConv, self).__init__()
        self._in_feats = in_feats
        self._out_feats = out_feats
        self._norm = True
        self._activation = activation


    def forward(self, graph, feat, weight, bias):

        graph = graph.local_var()
        if self._norm:
            norm = torch.pow(graph.in_degrees().float().clamp(min=1), -0.5)
            shp = norm.shape + (1,) * (feat.dim() - 1)
            norm = torch.reshape(norm, shp).to(feat.device)
            feat = feat * norm

        if self._in_feats > self._out_feats:
             
            feat = torch.matmul(feat, weight)
            graph.ndata['h'] = feat
            graph.update_all(fn.copy_src(src='h', out='m'),
                             fn.sum(msg='m', out='h'))
            rst = graph.ndata['h']
        else:
             
            graph.ndata['h'] = feat
            graph.update_all(fn.copy_src(src='h', out='m'),
                             fn.sum(msg='m', out='h'))
            rst = graph.ndata['h']
            rst = torch.matmul(rst, weight)

        rst = rst * norm

        rst = rst + bias

        if self._activation is not None:
            rst = self._activation(rst)

        return rst

    def extra_repr(self):
        """Set the extra representation of the module,
        which will come into effect when printing the model.
        """
        summary = 'in={_in_feats}, out={_out_feats}'
        summary += ', normalization={_norm}'
        if '_activation' in self.__dict__:
            summary += ', activation={_activation}'
        return summary.format(**self.__dict__)


class Classifier(nn.Module):
    def __init__(self, config, input_dim=512, mlp_dim=512, output_dim=512):
        super(Classifier, self).__init__()
        
        self.vars = nn.ParameterList()
        self.graph_conv = []
        self.config = config
        self.LinkPred_mode = False
        
        if self.config[-1][0] == 'LinkPred':
            self.LinkPred_mode = True
        
        for i, (name, param) in enumerate(self.config):
            
            if name is 'Linear':
                if self.LinkPred_mode:
                    w = nn.Parameter(torch.ones(param[1], param[0] * 2))
                else:
                    w = nn.Parameter(torch.ones(param[1], param[0]))
                init.kaiming_normal_(w)
                self.vars.append(w)
                self.vars.append(nn.Parameter(torch.zeros(param[1])))
            if name is 'GraphConv':
                 
                w = nn.Parameter(torch.Tensor(param[0], param[1]))
                init.xavier_uniform_(w)
                self.vars.append(w)
                self.vars.append(nn.Parameter(torch.zeros(param[1])))
                self.graph_conv.append(GraphConv(param[0], param[1], activation = F.relu))
            if name is 'Attention':
                 
                 
                 
                 
                 
                if self.LinkPred_mode:
                    w_q = nn.Parameter(torch.ones(param[1], param[0] * 2))
                else:
                    w_q = nn.Parameter(torch.ones(param[1], param[0]))
                w_k = nn.Parameter(torch.ones(param[1], param[0]))    
                w_v = nn.Parameter(torch.ones(param[1], param[4]))
                
                if self.LinkPred_mode:
                    w_l = nn.Parameter(torch.ones(param[3], param[2] * 2 + param[1]))
                else:
                    w_l = nn.Parameter(torch.ones(param[3], param[2] + param[1]))
                    
                init.kaiming_normal_(w_q)
                init.kaiming_normal_(w_k)
                init.kaiming_normal_(w_v)
                init.kaiming_normal_(w_l)

                self.vars.append(w_q)
                self.vars.append(w_k)
                self.vars.append(w_v)
                self.vars.append(w_l)

                 
                self.vars.append(nn.Parameter(torch.zeros(param[1])))
                self.vars.append(nn.Parameter(torch.zeros(param[1])))
                self.vars.append(nn.Parameter(torch.zeros(param[1])))
                 
                self.vars.append(nn.Parameter(torch.zeros(param[3])))

        self._build_mlp(2, input_dim, mlp_dim, output_dim, False)

    def forward(self, g, to_fetch, features, vars = None):
         
         

        if vars is None:
            vars = self.vars

        idx = 0 
        idx_gcn = 0
         
        h = features.float()
        h = h.to(device)

        for name, param in self.config:
            if name == 'GraphConv':
                w, b = vars[idx], vars[idx + 1]
                conv = self.graph_conv[idx_gcn]
                
                h = conv(g, h, w, b)

                g.ndata['h'] = h

                idx += 2 
                idx_gcn += 1

                if idx_gcn == len(self.graph_conv):
                     
                    num_nodes_ = g.batch_num_nodes().tolist()
                    temp = [0]+num_nodes_
                    offset = torch.cumsum(torch.LongTensor(temp), dim = 0)[:-1].to(device)
                    
                    if self.LinkPred_mode:
                        h1 = h[to_fetch[:,0] + offset]
                        h2 = h[to_fetch[:,1] + offset]
                        h = torch.cat((h1, h2), 1)
                    else:
                        h = h[to_fetch + offset]
                        
            if name == 'Linear':
                w, b = vars[idx], vars[idx + 1]
                h = F.linear(h, w, b)
                idx += 2

            if name == 'Attention':
                w_q, w_k, w_v, w_l = vars[idx], vars[idx + 1], vars[idx + 2], vars[idx + 3]
                b_q, b_k, b_v, b_l = vars[idx + 4], vars[idx + 5], vars[idx + 6], vars[idx + 7]

                Q = F.linear(h, w_q, b_q)
                K = F.linear(h, w_k, b_k)

                attention_scores = torch.matmul(Q, K.T)
                attention_probs = nn.Softmax(dim=-1)(attention_scores)
                context = F.linear(attention_probs, w_v, b_v)
                
                 
                 
                h = torch.cat((context, h), 1)
                h = F.linear(h, w_l, b_l)
                idx += 8
       
        return h, h

    def _build_mlp(self, n_layers, input_dim, mlp_dim, output_dim, last_bn=True):
        mlp = []
        for l in range(n_layers):
            dim1 = input_dim if l == 0 else mlp_dim
            dim2 = output_dim if l == n_layers - 1 else mlp_dim
        
            mlp.append(nn.Linear(dim1, dim2, bias=False))
        
            if l < n_layers - 1:
                mlp.append(nn.BatchNorm1d(dim2))
                mlp.append(nn.ReLU(inplace=True))
            elif last_bn:
                 
                 
                mlp.append(nn.BatchNorm1d(dim2, affine=False))
    
        return nn.Sequential(*mlp)
    
    def zero_grad(self, vars=None):

        with torch.no_grad():
            if vars is None:
                for p in self.vars:
                    if p.grad is not None:
                        p.grad.zero_()
            else:
                for p in vars:
                    if p.grad is not None:
                        p.grad.zero_()

    def parameters(self):
        return self.vars

class GAT(nn.Module):
    def __init__(self, n_layers, in_dim, n_hidden, n_classes, heads, activation, feat_drop, attn_drop, negative_slope, residual, norm=False, LinkPred_mode=False, maml=False):
        super(GAT, self).__init__()
        self.n_layers = n_layers
        self.gat_layers = nn.ModuleList()
        self.activation = activation
         
        self.gat_layers.append(GATConv(
            in_dim, n_hidden, heads[0],
            feat_drop, attn_drop, negative_slope, False, self.activation))
         
        for l in range(1, n_layers):
             
            self.gat_layers.append(GATConv(
                n_hidden * heads[l-1], n_hidden, heads[l],
                feat_drop, attn_drop, negative_slope, residual, self.activation))
         
        self.gat_layers.append(GATConv(
            n_hidden * heads[-2], n_classes, heads[-1],
            feat_drop, attn_drop, negative_slope, residual, None))

         
        self.readout = AvgPooling()
        self.norm = norm
        self.LinkPred_mode = LinkPred_mode
        self.maml = maml
    
    def forward(self, g, inputs, to_fetch=None):
        if self.maml:
            return self.finetunning_ProtoMAML_forward(g, inputs, to_fetch)
        else:
            return self.cl_forward(g, inputs)
    
    def cl_forward(self, g, inputs):
        h = inputs
        for l in range(self.n_layers):
            h = self.gat_layers[l](g, h).flatten(1)
         
         
         
        g.ndata['h'] = self.gat_layers[-1](g, h)

         
        overall_feat = self.readout(g, g.ndata['h'])
        if self.norm:
            overall_feat = F.normalize(overall_feat, p=2, dim=-1, eps=1e-5)
        return overall_feat  

    def finetunning_ProtoMAML_forward(self, g, inputs, to_fetch):
         
         

        h = inputs
        for l in range(self.n_layers):
            h = self.gat_layers[l](g, h).flatten(1)
         
         
         
        g.ndata['h'] = self.gat_layers[-1](g, h)

        num_nodes_ = g.batch_num_nodes().tolist()
        temp = [0] + num_nodes_
        offset = torch.cumsum(torch.LongTensor(temp), dim=0)[:-1].to(device)

        if self.LinkPred_mode:
            h1 = h[to_fetch[:, 0] + offset]
            h2 = h[to_fetch[:, 1] + offset]
            h = torch.cat((h1, h2), 1)
        else:
            h = h[to_fetch + offset]
    
        return h


class gat_net(GAT):
     
    def __init__(self, hidden_mlp_dim, out_mlp_dim,
                 n_layers, in_dim, n_hidden, n_classes, heads, activation, feat_drop, attn_drop, negative_slope, residual, norm, maml=False):
        super(gat_net, self).__init__(n_layers, in_dim, n_hidden, n_classes, heads, activation, feat_drop, attn_drop, negative_slope, residual, norm, maml=maml)
         
        self.fc = nn.Sequential(nn.Linear(n_classes, hidden_mlp_dim, bias=False), nn.ReLU(inplace=True), nn.BatchNorm1d(hidden_mlp_dim))
        self.predictor = nn.Sequential(nn.Linear(hidden_mlp_dim, out_mlp_dim, bias=False), nn.BatchNorm1d(out_mlp_dim, affine=True))
         
